@@ -109,6 +109,8 @@ Decoder_dealloc(DecoderObject *self)
         STAILQ_REMOVE_HEAD(&self->pending_blocks, entries);
         header_block_free(hblock);
     }
+
+    Py_TYPE(self)->tp_free((PyObject *) self);
 }
 
 static PyObject*
@@ -133,6 +135,7 @@ Decoder_feed_encoder(DecoderObject *self, PyObject *args, PyObject *kwargs)
         if (!hblock->blocked) {
             value = PyLong_FromUnsignedLongLong(hblock->stream_id);
             PyList_Append(list, value);
+            Py_DECREF(value);
         }
     }
     return list;
@@ -153,7 +156,7 @@ Decoder_feed_header(DecoderObject *self, PyObject *args, PyObject *kwargs)
     uint64_t stream_id;
     const unsigned char *data;
     Py_ssize_t data_len;
-    PyObject *list;
+    PyObject *control, *headers, *tuple;
     size_t dec_len = DEC_BUF_SZ;
     enum lsqpack_read_header_status status;
     struct header_block *hblock;
@@ -193,19 +196,20 @@ Decoder_feed_header(DecoderObject *self, PyObject *args, PyObject *kwargs)
         return NULL;
     }
 
-    list = hlist_to_headers(hblock->hlist);
+    control = PyBytes_FromStringAndSize((const char*)self->dec_buf, dec_len);
+    headers = hlist_to_headers(hblock->hlist);
     header_block_free(hblock);
 
-    return PyTuple_Pack(
-        2,
-        PyBytes_FromStringAndSize((const char*)self->dec_buf, dec_len),
-        list
-    );
+    tuple = PyTuple_Pack(2, control, headers);
+    Py_DECREF(control);
+    Py_DECREF(headers);
+
+    return tuple;
 }
 
 PyDoc_STRVAR(Decoder_feed_header__doc__,
-    "feed_header(stream_id: int, data: bytes) -> List[Tuple[bytes, bytes]]\n\n"
-    "Decode a header block and return headers.\n\n"
+    "feed_header(stream_id: int, data: bytes) -> Tuple[bytes, List[Tuple[bytes, bytes]]]\n\n"
+    "Decode a header block and return control data and headers.\n\n"
     "If the stream is blocked, :class:`StreamBlocked` is raised.\n\n"
     "If the data cannot be processed, :class:`DecompressionFailed` is raised.\n\n"
     ":param stream_id: the ID of the stream\n"
@@ -216,7 +220,7 @@ Decoder_resume_header(DecoderObject *self, PyObject *args, PyObject *kwargs)
 {
     char *kwlist[] = {"stream_id", NULL};
     uint64_t stream_id;
-    PyObject *list;
+    PyObject *control, *headers, *tuple;
     size_t dec_len = DEC_BUF_SZ;
     enum lsqpack_read_header_status status;
     struct header_block *hblock;
@@ -256,26 +260,27 @@ Decoder_resume_header(DecoderObject *self, PyObject *args, PyObject *kwargs)
         PyErr_Format(StreamBlocked, "stream %d is blocked", stream_id);
         return NULL;
     } else if (status != LQRHS_DONE) {
-        PyErr_Format(DecompressionFailed, "lsqpack_dec_header_in for stream %d failed (%d)", stream_id, status);
+        PyErr_Format(DecompressionFailed, "lsqpack_dec_header_read for stream %d failed (%d)", stream_id, status);
         STAILQ_REMOVE(&self->pending_blocks, hblock, header_block, entries);
         header_block_free(hblock);
         return NULL;
     }
 
-    list = hlist_to_headers(hblock->hlist);
+    control = PyBytes_FromStringAndSize((const char*)self->dec_buf, dec_len);
+    headers = hlist_to_headers(hblock->hlist);
     STAILQ_REMOVE(&self->pending_blocks, hblock, header_block, entries);
     header_block_free(hblock);
 
-    return PyTuple_Pack(
-        2,
-        PyBytes_FromStringAndSize((const char*)self->dec_buf, dec_len),
-        list
-    );
+    tuple = PyTuple_Pack(2, control, headers);
+    Py_DECREF(control);
+    Py_DECREF(headers);
+
+    return tuple;
 }
 
 PyDoc_STRVAR(Decoder_resume_header__doc__,
-    "resume_header(stream_id: int) -> List[Tuple[bytes, bytes]]\n\n"
-    "Continue decoding a header block and return headers.\n\n"
+    "resume_header(stream_id: int) -> Tuple[bytes, List[Tuple[bytes, bytes]]]\n\n"
+    "Continue decoding a header block and return control data and headers.\n\n"
     "This method should be called only when :meth:`feed_encoder` indicates "
     "that a stream has become unblocked\n\n"
     ":param stream_id: the ID of the stream\n");
@@ -354,6 +359,7 @@ static void
 Encoder_dealloc(EncoderObject *self)
 {
     lsqpack_enc_cleanup(&self->enc);
+    Py_TYPE(self)->tp_free((PyObject *) self);
 }
 
 static PyObject*
@@ -441,11 +447,13 @@ Encoder_encode(EncoderObject *self, PyObject *args, PyObject *kwargs)
     pfx_off = PREFIX_MAX_SIZE - pfx_len;
     memcpy(self->hdr_buf + pfx_off, self->pfx_buf, pfx_len);
 
-    return PyTuple_Pack(
-        2,
-        PyBytes_FromStringAndSize((const char*)self->enc_buf, enc_off),
-        PyBytes_FromStringAndSize((const char*)self->hdr_buf + pfx_off, hdr_off - pfx_off)
-    );
+    name = PyBytes_FromStringAndSize((const char*)self->enc_buf, enc_off);
+    value = PyBytes_FromStringAndSize((const char*)self->hdr_buf + pfx_off, hdr_off - pfx_off);
+    tuple = PyTuple_Pack(2, name, value);
+    Py_DECREF(name);
+    Py_DECREF(value);
+
+    return tuple;
 }
 
 PyDoc_STRVAR(Encoder_encode__doc__,
